#ifndef __GENERATED_MEM_H
#define __GENERATED_MEM_H

#define VEXRISCV_DEBUG_BASE 0xf00f0000
#define VEXRISCV_DEBUG_SIZE 0x00000010

#define SRAM_BASE 0x10000000
#define SRAM_SIZE 0x00020000

#define ROM_BASE 0x00000000
#define ROM_SIZE 0x00002000

#define SPIFLASH_BASE 0x20000000
#define SPIFLASH_SIZE 0x00200000

#endif
