#ifndef __HW_FLAGS_H
#define __HW_FLAGS_H

#define UART_EV_TX	0x1
#define UART_EV_RX	0x2

#define DFII_CONTROL_SEL		0x01
#define DFII_CONTROL_CKE		0x02
#define DFII_CONTROL_ODT		0x04
#define DFII_CONTROL_RESET_N	0x08

#define DFII_COMMAND_CS		0x01
#define DFII_COMMAND_WE		0x02
#define DFII_COMMAND_CAS	0x04
#define DFII_COMMAND_RAS	0x08
#define DFII_COMMAND_WRDATA	0x10
#define DFII_COMMAND_RDDATA	0x20

#define ETHMAC_EV_SRAM_WRITER	0x1
#define ETHMAC_EV_SRAM_READER	0x1

#define CLKGEN_STATUS_BUSY		0x1
#define CLKGEN_STATUS_PROGDONE	0x2
#define CLKGEN_STATUS_LOCKED	0x4

#define DVISAMPLER_TOO_LATE		0x1
#define DVISAMPLER_TOO_EARLY	0x2

#define DVISAMPLER_DELAY_MASTER_CAL	0x01
#define DVISAMPLER_DELAY_MASTER_RST	0x02
#define DVISAMPLER_DELAY_SLAVE_CAL	0x04
#define DVISAMPLER_DELAY_SLAVE_RST	0x08
#define DVISAMPLER_DELAY_INC		0x10
#define DVISAMPLER_DELAY_DEC		0x20

#define DVISAMPLER_SLOT_EMPTY	0
#define DVISAMPLER_SLOT_LOADED	1
#define DVISAMPLER_SLOT_PENDING	2

#endif /* __HW_FLAGS_H */
